from functools import wraps

from furthrmind.utils import furthr_wrap, instance_overload
from typing_extensions import List, Self, Any, Dict, TYPE_CHECKING
from inspect import isclass
import os

if TYPE_CHECKING:
    from furthrmind.collection import File, FieldData, Experiment, Sample, ResearchItem

class BaseClass:
    _data = {}
    _fetched = False
    _id = None
    fm = None

    _attr_definition = {}

    def __init__(self, id=None, data=None):
        if data:
            self._update_attributes(data)

        if id:
            self._id = id
        else:
            if "id" in self._data:
                self._id = self._data["id"]
        if not self._id:
            raise ValueError("No id provided")

        # create instance methods for certain class_methods
        instance_methods = ["get", "get_all", "post"]
        instance_overload(self, instance_methods)

    def __getitem__(self, item):
        if hasattr(self, item):
            return getattr(self, item)
        else:
            raise ValueError("No such attribute")


    def _get_url_instance(self, project_id=None):
        return ""

    @classmethod
    def _get_url_class(cls, id, project_id=None):
        return ""

    @classmethod
    def _get_all_url(cls, project_id=None):
        return ""

    @classmethod
    def _post_url(cls, project_id=None):
        return ""

    @staticmethod
    def _update_instance_decorator(f):
        @wraps(f)
        def decorated(*args, **kws):
            results = f(*args, **kws)
            self = args[0]
            self._update_attributes(results)
            return self

        return decorated

    @staticmethod
    def _create_instances_decorator(f):
        @wraps(f)
        def decorated(*args, **kws):
            results = f(*args, **kws)
            if results is None:
                return
            if isclass(args[0]):
                cls = args[0]
            else:
                self = args[0]
                cls = self.__class__

            if isinstance(results, list):
                item_list = []
                for r in results:
                    item = cls(data=r)
                    item_list.append(item)
                return item_list
            else:
                item = cls(data=results)
                return item

        return decorated

    def _update_attributes(self, data):
        from furthrmind.collection import get_collection_class
        self._data = data

        def _create_instance(classname, _data):
            cls = get_collection_class(classname)
            if isinstance(_data, list):
                item_list = []
                for item in _data:
                    item = cls(self.fm, data=item)
                    item_list.append(item)
                item = item_list
            else:
                item = cls(self.fm, data=_data)
            return item

        for key in data:
            if hasattr(self, key):
                item = data[key]
                if key in self._attr_definition:
                    attr_definition = self._attr_definition[key]
                    if "class" in attr_definition:
                        if "nested_dict" in attr_definition:
                            item = {}
                            for item_key in data[key]:
                                item[item_key] = _create_instance(attr_definition["class"],
                                                                  data[key][item_key])
                        else:
                            if data[key]:
                                item = _create_instance(attr_definition["class"], data[key])
                            else:
                                pass

                setattr(self, key, item)
            else:
                for attr_definition in self._attr_definition.items():
                    definition_key = attr_definition[0]
                    definition_value = attr_definition[1]
                    if "data_key" in definition_value:
                        if definition_value["data_key"] == key:
                            item = data[key]
                            if "class" in definition_value:
                                item = _create_instance(definition_value["class"], data[key])
                            setattr(self, definition_key, item)
                            break

    @classmethod
    def get(cls, id=None):
        if isclass(cls):
            return cls._get_class_method(id)
        else:
            self = cls
            data = self._get_instance_method()
            return data

    @_update_instance_decorator
    @furthr_wrap(force_list=False)
    def _get_instance_method(self):
        url = self._get_url_instance()
        data = self.fm.session.get(url)
        return data

    @classmethod
    @_create_instances_decorator
    @furthr_wrap(force_list=False)
    def _get_class_method(cls, id):
        url = cls._get_url_class(id)
        return cls.fm.session.get(url)

    @classmethod
    def get_all(cls, project_id=None) -> List[Self]:
        if isclass(cls):
            return cls._get_all_class_method(project_id)
        else:
            self = cls
            return self._get_all_instance_method(project_id)

    @_create_instances_decorator
    @furthr_wrap(force_list=True)
    def _get_all_instance_method(self, project_id):
        from .project import Project
        if isinstance(self, Project):
            url = self.__class__._get_all_url()
        else:
            url = self.__class__._get_all_url(project_id)
        return self.fm.session.get(url)

    @classmethod
    @_create_instances_decorator
    @furthr_wrap(force_list=True)
    def _get_all_class_method(cls, project_id):
        from .project import Project
        if cls in [Project]:
            url = cls._get_all_url()
        else:
            url = cls._get_all_url(project_id)
        return BaseClass.fm.session.get(url)
      

    @classmethod
    def post(cls, data, project_id=None):
        if isclass(cls):
            return cls._post_class_method(data, project_id)
        else:
            self = cls
            return self._post_instance_method(data, project_id)


    @furthr_wrap(force_list=False)
    def _post_instance_method(self, data,  project_id=None):
        url = self.__class__._post_url(project_id)
        return self.fm.session.post(url, json=data)

    @classmethod
    @furthr_wrap(force_list=False)
    def _post_class_method(cls, data, project_id=None):
        url = cls._post_url(project_id)
        return cls.fm.session.post(url, json=data)

    def to_dict(self):
        from furthrmind import Furthrmind
        data = {}
        for attr in dir(self):
            if attr.startswith('_'):
                continue
            value = getattr(self, attr)
            if callable(value):
                continue
            if isinstance(value, Furthrmind):
                continue
            data[attr] = self._convert(value)
        return data

    def _convert(self, item):
        if isinstance(item, dict):
            new_item = {}
            for key, value in item.items():
                if isinstance(value, (dict, list)):
                    value = self._convert(value)
                    new_item[key] = value
                elif isinstance(value, BaseClass):
                    value = value.to_dict()
                    new_item[key] = value
                else:
                    new_item[key] = value
        elif isinstance(item, list):
            new_item = []
            for value in item:
                if isinstance(value, (dict, list)):
                    value = self._convert(value)
                    new_item.append(value)
                elif isinstance(value, BaseClass):
                    value = value.to_dict()
                    new_item.append(value)
                else:
                    new_item.append(value)
        elif isinstance(item, BaseClass):
            new_item = item.to_dict()

        else:
            new_item = item

        return new_item

class BaseClassWithFieldData(BaseClass):
    id = None
    fielddata: List["FieldData"] = []


    def __init__(self, id=None, data=None):
        super().__init__(id, data)

    def update_field_value(self, value, field_name=None, field_id=None):
        """
        :param value:
            - Numeric: float or int, or a string convertable to a float
            - Date: datetime, or date object, or unix timestamp or string with iso format
            - SingleLine: string
            - ComboBoxEntry: dict with id or name as key, or string with name, or string with id
            - MultiLine: dict with content as key, or string
            - CheckBox: boolean
        :param field_name: Name of field that should be updated. Either field_name or field_id must be specified
        :param field_id: id of field that should be updated. Either field_name or field_id must be specified
        :return: id
        """

        fielddata = None
        for item in self.fielddata:
            if fielddata:
                break
            if field_id:
                if item.field_id == field_id:
                    fielddata = item
            elif field_name:
                if item.field_name == field_name:
                    fielddata = item

        if not fielddata:
            raise ValueError("No field found with the given field_id or field_name")

        return fielddata.update_value(value)

    def update_field_unit(self, unit, field_name=None, field_id=None):
        """
        :param value:
            - Numeric: float or int, or a string convertable to a float
            - Date: datetime, or date object, or unix timestamp or string with iso format
            - SingleLine: string
            - ComboBoxEntry: dict with id or name as key, or string with name, or string with id
            - MultiLine: dict with content as key, or string
            - CheckBox: boolean
        :param field_name: Name of field that should be updated. Either field_name or field_id must be specified
        :param field_id: id of field that should be updated. Either field_name or field_id must be specified
        :return: id
        """

        fielddata = None
        for item in self.fielddata:
            if fielddata is not None:
                break
            if field_id:
                if item.fieldid == field_id:
                    fielddata = item
            elif field_name:
                if item.field_name == field_name:
                    fielddata = item

        if not fielddata:
            raise ValueError("No field found with the given field_id or field_name")

        return fielddata.update_unit(unit)

    def set_calculation_result(self, value: dict, field_name=None, field_id=None, fielddata_id=None):
        """
        :param value
        :param field_name: Name of field that should be updated. Either field_name, field_id, fielddata_id must be specified
        :param field_id: id of field that should be updated. Either field_name, field_id, or fielddata_id must be specified
        :param field_data_id: id of the fielddata that should be updated. Either field_name, field_id, or fielddata_id must be specified
        :return: id
        """

        fielddata = None
        for item in self.fielddata:
            if fielddata:
                break
            if fielddata_id:
                if item.id == fielddata_id:
                    fielddata = item
            if field_id:
                if item.fieldid == field_id:
                    fielddata = item
            elif field_name:
                if item.field_name == field_name:
                    fielddata = item

        if not fielddata:
            raise ValueError("No field found with the given field_id or field_name")

        return fielddata.set_calculation_result(value)

    def add_field(self, field_name=None, field_type=None, field_id=None,
                  value=None, unit=None) -> "FieldData":
        """
        :param field_name: Name of field that should be added. If fieldname provided,
                            also fieldtype must be specified. Either fieldname and fieldtype or field_id must be specified
        :param field_type: Type of field: Must be out of: Numeric, Date, SingleLine
                                                        ComboBoxEntry, MultiLine, CheckBox
        :param field_id: id of field that should be added.
        :param value:
            - Numeric: float or int, or a string convertable to a float
            - Date: datetime, or date object, or unix timestamp or string with iso format
            - SingleLine: string
            - ComboBoxEntry: dict with id or name as key, or string with name, or string with id
            - MultiLine: dict with content as key, or string
            - CheckBox: boolean
        :param unit: dict with id or name, or name as string, or id as string
        :return: fielddata object
        """

        from .fielddata import FieldData

        fielddata = FieldData.create(field_name, field_type, field_id, value, unit)
        self.fielddata.append(fielddata)
        data = {"id": self.id, "fielddata": [{"id": f.id} for f in self.fielddata]}
        self.post(data)
        return fielddata

    def add_many_fields(self, data_list: List[Dict] ) -> List["FieldData"]:
        """
        Method to add many fields to an item. Each field is defined by and dict in the data_list parameter

        :param data_list: dict with the following key
        - field_name: Name of field that should be added. If fieldname provided, also fieldtype must be specified. Either fieldname and fieldtype or field_id must be specified
        - field_type: Type of field: Must be out of: Numeric, Date, SingleLine
            ComboBoxEntry, MultiLine, CheckBox
        - field_id: id of field that should be added.
        - value:
            - Numeric: float or int, or a string convertable to a float
            - Date: datetime, or date object, or unix timestamp or string with iso format
            - SingleLine: string
            - ComboBoxEntry: dict with id or name as key, or string with name, or string with id
            - MultiLine: dict with content as key, or string
            - CheckBox: boolean
        - unit: dict with id or name, or name as string, or id as string

        :return: list with fielddata object

        """

        from .fielddata import FieldData

        fielddata_list = FieldData.create_many(data_list)
        self.fielddata.extend(fielddata_list)
        data = {"id": self.id, "fielddata": [{"id": f.id} for f in self.fielddata]}
        self.post(data)
        return fielddata_list


    def remove_field(self, fieldname=None, fieldid=None):
        """

        :param fieldname: Name of field that should be removed. Either id or fieldname must be specified
        :param fieldid: id of field that should be removed.
        :return id of item

        """

        new_fielddata_list = []
        fielddata_to_be_removed = None
        for fielddata in self.fielddata:
            found = False
            if fieldid:
                found = True
                if fielddata.fieldid == fieldid:
                    fielddata_to_be_removed = fielddata
            elif fieldname:
                found = True
                if fielddata.field_name == fieldname:
                    fielddata_to_be_removed = fielddata
            if not found:
                new_fielddata_list.append(fielddata)


        if not fielddata_to_be_removed:
            raise ValueError("No field found with the given fieldid or fieldname")

        self.fielddata = new_fielddata_list
        fielddata_list = [{"id": fd.id} for fd in new_fielddata_list]
        post_data = {"id": self.id, "fielddata": fielddata_list}
        id = self.post(post_data)
        return id

class BaseClassWithFiles(BaseClass):
    id = None
    files = []

    def __init__(self, id=None, data=None):
        super().__init__(id, data)

    def add_file(self, file_path, file_name=None) -> "File":
        """

        :param file_path: file path of the file that should be uploaded
        :param file_name: Optionally specify the file name if not the original file name should be used

        :return: file object
        """

        from furthrmind.file_loader import FileLoader
        from .file import File

        if not os.path.isfile(file_path):
            raise ValueError("File does not exist")

        fl = FileLoader(self.fm.host, self.fm.api_key)
        file_id, file_name = fl.uploadFile(file_path, file_name)
        file_data = {"id": file_id,
                     "name": file_name}
        file_list = [{"id": f.id} for f in self.files]
        file_list.append(file_data)
        post_data = {"id": self.id, "files": file_list}

        id = self.post(post_data)
        file = File(data=file_data)
        self.files.append(file)
        return file

    def remove_file(self, file_id=None, file_name=None):
        """

        :param file_id: id of file that should be removed, either file_name or file_id must be specified
        :param file_name: file name of file to be removed

        :return: file object
        """

        new_file_list = []
        file_to_be_removed = None
        for file in self.files:
            found = False
            if file_id:
                found = True
                if file.id == file_id:
                    file_to_be_removed = file
            elif file_name:
                found = True
                if file.name == file_name:
                    file_to_be_removed = file
            if not found:
                new_file_list.append(file)

        if not file_to_be_removed:
            raise ValueError("No file found with the given file_id or file_name")

        self.files = new_file_list
        file_list = [{"id": f.id} for f in new_file_list]
        post_data = {"id": self.id, "files": file_list}
        id = self.post(post_data)
        return id

class BaseClassWithGroup(BaseClass):
    id = None
    groups = []

    def __init__(self, id=None, data=None):
        super().__init__(id, data)

    @classmethod
    def _create(cls, name, group_name = None, group_id=None, project_id=None):
        """
        Method to create a new item (sample, experiment)
        :param name: the name of the item to be created
        :param group_name: The name of the group where the new item will belong to. group name can be only considered
                            for groups that are not subgroups. Either group_name or group_id must be specified
        : param group_id: the id of the group where the new item will belong to. Either group_name or group_id must be specified
        : param project_id: Optionally to create an item in another project as the furthrmind sdk was initiated with
        : return: instance of the item class: experiment, or sample
        """
        data = cls._prepare_data_for_create(name, group_name, group_id, project_id)
        id = cls.post(data, project_id)
        data["id"] = id
        return data

    @classmethod
    def _create_many(cls, data_list: List[Dict], project_id=None):
        """
        Method to create multiple items (sample, experiment)
        :param data_list: dict with the following keys:
            - name: the name of the item to be created
            - group_name: The name of the group where the new item will belong to. group name can be only considered
                            for groups that are not subgroups. Either group_name or group_id must be specified
            - group_id: the id of the group where the new item will belong to. Either group_name or group_id must be specified
        :param project_id: Optionally to create an item in another project as the furthrmind sdk was initiated with
        : return list with instances of the item class: experiment, or sample
        """
        from furthrmind.collection import Group

        new_list = []
        groups = Group.get_all()
        for data in data_list:
            new_list.append(cls._prepare_data_for_create(name=data.get("name"), group_name=data.get("group_name"),
                                                         group_id=data.get("group_id"), project_id=project_id,
                                                         groups=groups))

        id_list = cls.post(new_list, project_id)
        for data, id in zip(new_list, id_list):
            data["id"] = id
        return new_list

    @classmethod
    def _prepare_data_for_create(cls, name, group_name = None, group_id=None, project_id=None, groups=None):
        from furthrmind.collection import Group
        if not name:
            raise ValueError("Name must be specified")
        if not group_name and not group_id:
            raise ValueError("Either group_name or group_id must be specified")

        if group_name:
            if not groups:
                groups = Group.get_all(project_id)
            for group in groups:
                if group.parent_group:
                    continue
                if group.name == group_name:
                    group_id = group.id
                    break

            if not group_id:
                raise ValueError("No group with Name was found")

        data = {"name": name, "groups": [{"id": group_id}]}

        return data

class BaseClassWithLinking(BaseClass):
    id: str = ""
    linked_samples: List["Sample"] = []
    linked_experiments: List["Experiment"] = []
    linked_researchitems: Dict[str, List["ResearchItem"]] = []

    def __init__(self, id=None, data=None):
        super().__init__(id, data)

    def add_linked_experiment(self, experiment_id=None, experiment_name=None):
        """
        This method is to link an experiment to the current item
        Args:
            experiment_id: id to the experiment you want to link, either id or name must be given
            experiment_name: name of the experiment you want to link, either name or name must be given

        Returns:
            the id of the item
        """
        from furthrmind.collection import Experiment
        assert experiment_id or experiment_name, "Either experiment_id or experiment_name must be specified"

        if experiment_name:
            exp = Experiment.get(name = experiment_name)
            if not exp:
                raise ValueError("No exp found with the given name")
            experiment_id = exp.id
        else:
            exp = Experiment.get(experiment_id)

        experiment_id_list = [item.id for item in self.linked_experiments]
        if experiment_id not in experiment_id_list:
            experiment_id_list.append(experiment_id)

        linked_experiment = [{"id": exp_id} for exp_id in experiment_id_list]
        data = {
            "id": self.id,
                "experiments": linked_experiment
        }

        self.post(data=data)
        self.linked_experiments.append(exp)
        return self.id

    def remove_linked_experiment(self, experiment_id=None, experiment_name=None):
        """
        This method is to remove link a linked experiment from the current item
        Args:
            experiment_id: id to the experiment you want to remove the linkage, either id or name must be given
            experiment_name: name of the experiment you want to remove the linkage, either name or name must be given

        Returns:
            the id of the item
        """
        from furthrmind.collection import Experiment
        assert experiment_id or experiment_name, "Either experiment_id or experiment_name must be specified"

        if experiment_name:
            exp = Experiment.get(name = experiment_name)
            if not exp:
                raise ValueError("No exp found with the given name")
            experiment_id = exp.id

        experiment_id_list = []
        new_linked_items = []
        for item in self.linked_experiments:
            if item.id == experiment_id:
                continue
            new_linked_items.append(item)
            experiment_id_list.append(item.id)

        linked_experiment = [{"id": exp_id} for exp_id in experiment_id_list]
        data = {
            "id": self.id,
                "experiments": linked_experiment
        }

        self.post(data=data)
        self.linked_experiments = new_linked_items
        return self.id

    def add_linked_sample(self, sample_id=None, sample_name=None):
        """
        This method is to link a sample to the current item
        Args:
            sample_id: id to the sample you want to link, either id or name must be given
            sample_name: name of the sample you want to link, either name or name must be given

        Returns:
            the id of the item
        """
        from furthrmind.collection import Sample
        assert sample_id or sample_name, "Either sample_id or sample_name must be specified"

        if sample_name:
            s = Sample.get(name=sample_name)
            if not s:
                raise ValueError("No sample found with the given name")
            sample_id = s.id
        else:
            s = Sample.get(sample_id)

        sample_id_list = [item.id for item in self.linked_samples]
        if sample_id not in sample_id_list:
            sample_id_list.append(sample_id)

        linked_samples = [{"id": s_id} for s_id in sample_id_list]

        data = {
            "id": self.id,
            "samples": linked_samples
        }

        self.post(data=data)
        self.linked_samples.append(s)
        return self.id

    def remove_linked_sample(self, sample_id=None, sample_name=None):
        """
        This method is to remove a linked sample from the current item
        Args:
            sample_id: id of the sample you want to remove the linkage, either id or name must be given
            sample_name: name of the sample you want to remove linkage, either name or name must be given

        Returns:
            the id of the item
        """
        from furthrmind.collection import Sample
        assert sample_id or sample_name, "Either sample_id or sample_name must be specified"

        if sample_name:
            s = Sample.get(name=sample_name)
            if not s:
                raise ValueError("No sample found with the given name")
            sample_id = s.id

        sample_id_list = []
        new_linked_items = []
        for item in self.linked_samples:
            if item.id == sample_id:
                continue
            new_linked_items.append(item)
            sample_id_list.append(item.id)

        linked_samples = [{"id": s_id} for s_id in sample_id_list]

        data = {
            "id": self.id,
            "samples": linked_samples
        }
        self.post(data=data)
        self.linked_samples = new_linked_items
        return self.id

    def add_linked_researchitem(self, researchitem_id=None):
        """
        This method is to link an experiment to the current item
        Args:
            researchitem_id: id to the researchitem you want to link

        Returns:
            the id of the item
        """
        assert researchitem_id, "Either experiment_id or experiment_name must be specified"

        researchitem_id_list = []
        for cat in self.linked_researchitems:
            researchitem_id_list.extend([ri_id for ri_id in self.linked_researchitems[cat]])

        if researchitem_id not in researchitem_id_list:
            researchitem_id_list.append(researchitem_id)

        linked_researchitems = [{"id": ri_id} for ri_id in researchitem_id_list]


        data = {
            "id": self.id,
            "researchitems": linked_researchitems
        }

        self.post(data=data)
        ri = ResearchItem.get(id=researchitem_id)
        self.linked_researchitems[ri.category.name].append(ri)
        return self.id

    def remove_linked_researchitem(self, researchitem_id=None):
        """
        This method is to link an experiment to the current item
        Args:
            researchitem_id: id to the researchitem you want to remove the linkage

        Returns:
            the id of the item
        """
        assert researchitem_id, "Either experiment_id or experiment_name must be specified"

        researchitem_id_list = []
        new_linked_items = {}
        for cat in self.linked_researchitems:
            for item in self.linked_researchitems[cat]:
                if item.id == researchitem_id:
                    continue
                if cat not in new_linked_items:
                    new_linked_items[cat] = []
                new_linked_items[cat].append(item)
                researchitem_id_list.append(item.id)

        if researchitem_id in researchitem_id_list:
            researchitem_id_list.remove(researchitem_id)

        linked_researchitems = [{"id": ri_id} for ri_id in researchitem_id_list]

        data = {
            "id": self.id,
            "researchitems": linked_researchitems
        }

        self.post(data=data)
        self.linked_researchitems = new_linked_items
        return self.id



from furthrmind.collection.baseclass import BaseClassWithFieldData, BaseClass
from typing_extensions import List, Self, Dict, TYPE_CHECKING
from inspect import isclass
if TYPE_CHECKING:
    from furthrmind.collection import *

class ComboBoxEntry(BaseClassWithFieldData):
    id = ""
    name = ""
    files: List["File"] = []
    fielddata: List["FieldData"] = []

    _attr_definition = {
        "files": {"class": "File"},
        "fielddata": {"class": "FieldData"}
    }

    def __init__(self, id=None, data=None):
        super().__init__(id, data)

    def _get_url_instance(self, project_id=None):
        project_url = ComboBoxEntry.fm.get_project_url(project_id)
        url = f"{project_url}/comboboxentry/{self.id}"
        return url

    @classmethod
    def _get_url_class(cls, id, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/comboboxentry/{id}"
        return url

    @classmethod
    def _get_all_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/comboboxentry"
        return url

    @classmethod
    def _post_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/comboboxentry"
        return url

    @classmethod
    def get(cls, id=None) -> Self:
        """
        Method to get all one comboboxentry by it's id
        If called on an instance of the class, the id of the class is used
        :param str id: id of requested comboboxentry 
        :return Self: Instance of comboboxentry class
        """
        if isclass(cls):
            return cls._get_class_method(id)
        else:
            self = cls
            data = self._get_instance_method()
            return data
    
    @classmethod
    def get_all(cls, project_id=None) -> List[Self]:
        """
        Method to get all comboboxentries belonging to one project
        :param str project_id: Optionally to get comboboxentries from another project as the furthrmind sdk was initiated with, defaults to None
        :return List[Self]: List with instances of comboboxentry class
        """
        return super().get_all(project_id)
    
    @classmethod
    @BaseClass._create_instances_decorator
    def create(cls, name: str, field_name: str = None, field_id=None, project_id=None) -> Self:
        """
        Method to create a new combobox entry

        :param name: name of the combobox entry
        :param field_name: Name of the field, where the combobox entry should belong to. Either the field name or id
           must be provided
        :param field_id: id of the field, where the combobox entry should belong to. Either the field name or id must
            be provided
        :param project_id: Optionally to create an item in another project as the furthrmind sdk was initiated with
        :return: instance of column comboboxentry class

        """
        if not name:
            raise ValueError("Name must be specified")
        if not field_name and not field_id:
            raise ValueError("Either field_name or field_id must be provided")

        if field_name:
            fields = Field.get_all(project_id)
            for field in fields:
                if field.name == field_name:
                    field_id = field.id
                    break

            if not field_id:
                raise ValueError("Field with given name not found")

        data = {"name": name, "field": {"id": field_id}}
        id = cls.post(data, project_id)
        data["id"] = id
        return data

    @classmethod
    @BaseClass._create_instances_decorator
    def create_many(cls, data_list: List[Dict], project_id=None) -> Self:
        """
        Method to create a new data column

        :param data_list: dict with the following keys:
            - name of the combobox entry
            - field_name: Name of the field, where the combobox entry should belong to. Either the field name or id
            must be provided
            - field_id: id of the field, where the combobox entry should belong to. Either the field name or id must
            be provided
        :param project_id: Optionally to create an item in another project as the furthrmind sdk was initiated with
        :return: List with instances of comboboxentry class

        """

        look_for_field_ids = False
        for data in data_list:
            if not data.get("name"):
                raise ValueError("Name must be specified")

            if data.get("field_name"):
                look_for_field_ids = True
                break

        if look_for_field_ids:
            fields = Field.get_all(project_id)
            for data in data_list:
                field_name = data.get("field_name")
                field_id = data.get("field_id")
                if not field_name and not field_id:
                    raise ValueError("Either field_name or field_id must be provided")
                if field_name:
                    for field in fields:
                        if field.name == field_name:
                            field_id = field.id
                            data["field_id"] = field_id
                            break
                        if not data.get("field_id"):
                            raise ValueError(f"Field with given name '{field_name}' not found")

        new_data_list = []
        for data in data_list:
            new_data_list.append({
                "name": data.get("name"),
                "field": {"id": data.get("field_id")}
            })

        id_list = cls.post(new_data_list, project_id)
        for data, id in zip(new_data_list, id_list):
            data["id"] = id

        return new_data_list






from typing_extensions import Self, List, Dict
from inspect import isclass
from ..utils import furthr_wrap
from furthrmind.collection.baseclass import BaseClass
from furthrmind.collection.fielddata import FieldData


class Column(BaseClass):
    id = ""
    name = ""
    type = ""
    values = []


    _attr_definition = {
        "columns": {"class": "Column"}
    }

    def __init__(self, id=None, data=None):
        super().__init__(id, data)

    @classmethod
    def get(cls, id=None) -> Self:
        """
        Method to get all one column by it's id
        If called on an instance of the class, the id of the class is used
        :param str id: id of requested column 
        :return Self: Instance of column class
        """

        if isclass(cls):
            return cls._get_class_method(id)
        else:
            self = cls
            data = self._get_instance_method()
            return data
    
    @classmethod
    def get_all(cls, project_id=None) -> List[Self]:
        raise ValueError("Not implemented for datatables")

    @classmethod
    @BaseClass._create_instances_decorator
    @furthr_wrap(force_list=True)
    def get_many(cls, id_list, project_id=None) -> List[Self]:
        """
        Method to get many columns

        :param id_list: list of column ids that should be retrieved
        :param project_id: Optionally to create an item in another project as the furthrmind sdk was initiated with
        :return: list of column objects

        """
        column_id_string = ",".join(id_list)
        project_url = Column.fm.get_project_url(project_id)
        url = f"{project_url}/columns/{column_id_string}"
        columns = Column.fm.session.get(url)
        return columns

    def _get_url_instance(self, project_id=None):
        project_url = Column.fm.get_project_url(project_id)
        url = f"{project_url}/column/{self.id}"
        return url

    @classmethod
    def _get_url_class(cls, id, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/column/{id}"
        return url

    @classmethod
    def _post_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/column"
        return url

    @classmethod
    def _get_all_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/column"
        return url

    @classmethod
    def _type_check(cls, column_type, data):
        if not column_type in ["Text", "Numeric"]:
            raise ValueError("Column type must be Text/Numeric")

        if column_type == "Text":
            if all([isinstance(d, str) for d in data]):
                return data
            return [str(d) for d in data]

        elif column_type == "Numeric":
            if all([isinstance(d, (int, float)) for d in data]):
                return data
            new_data = []
            for d in data:
                try:
                    new_data.append(float(d))
                except:
                    raise ValueError("All column values must be a float, int or a string that can be converted to a float")
            return new_data

    @classmethod
    @BaseClass._create_instances_decorator
    def create(cls, name: str, type: str, data: list, unit=None, project_id=None) -> Self:
        """
        Method to create a new data column

        :param name: Name of the column
        :param type: Type of the column, Either "Text" or "Numeric". Data must fit to type, for Text all data
            will be converted to string and for Numeric all data is converted to float (if possible)
        :param data: List of column values, must fit to column_type
        :param unit: dict with id or name, or name as string, or id as string
        :param project_id: Optionally to create an item in another project as the furthrmind sdk was initiated with
        :return: Instance of column class

        """

        data = cls._type_check(type, data)
        unit = FieldData._check_unit(unit)
        data_dict = {"name": name, "type": type, "values": data, "unit": unit}
        id = cls.post(data_dict, project_id)
        data_dict["id"] = id
        return data_dict

    @classmethod
    @BaseClass._create_instances_decorator
    def create_many(cls, data_list: List[Dict], project_id=None) -> Self:
        """
        Method to create a new data column

        :param data_list: dict with the following keys:
            - name: Name of the column
            - type: Type of the column, Either "Text" or "Numeric". Data must fit to type, for Text all data
            will be converted to string and for Numeric all data is converted to float (if possible)
            - data: List of column values, must fit to column_type
            - unit: dict with id or name, or name as string, or id as string
        :param project_id: Optionally to create an item in another project as the furthrmind sdk was initiated with
        :return: List with instances of column class

        """

        new_data_list = []
        for item in data_list:
            type = item.get("type")
            unit = item.get("unit")
            name = item.get("name")
            data = item.get("data")
            data = cls._type_check(type, data)
            unit = FieldData._check_unit(unit)
            data_dict = {"name": name, "type": type, "values": data, "unit": unit}
            new_data_list.append(data_dict)

        id_list = cls.post(new_data_list, project_id)
        for item, id in zip(new_data_list, id_list):
            item["id"] = id

        return new_data_list


from furthrmind.collection.baseclass import BaseClass
from typing_extensions import List, Self, Dict, TYPE_CHECKING
from inspect import isclass
if TYPE_CHECKING:
    from furthrmind.collection.comboboxentry import ComboBoxEntry

class Field(BaseClass):
    id = ""
    name = ""
    type = ""
    script = ""
    comboboxentries: List["ComboBoxEntry"] = []

    _attr_definition = {
        "comboboxentries": {"class": "ComboBoxEntry"}
    }

    def __init__(self, id=None, data=None):
        super().__init__(id, data)

    def _get_url_instance(self, project_id=None):
        project_url = Field.fm.get_project_url(project_id)
        url = f"{project_url}/fields/{self.id}"
        return url

    @classmethod
    def _get_url_class(cls, id, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/fields/{id}"
        return url

    @classmethod
    def _get_all_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/fields"
        return url

    @classmethod
    def _post_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/fields"
        return url

    @classmethod
    def get(cls, id=None, name=None) -> Self:
        """
        Method to get all one field by it's id
        :param str id: id of requested field 
        :param str name: name of requested field 
        :return Self: Instance of field class
        """
        

        if isclass(cls):
            if id is None:
                id = name
            assert id is not None or name is not None, "Either id or name must be specified"
            return cls._get_class_method(id)
        else:
            self = cls
            data = self._get_instance_method()
            return data
    
    @classmethod
    def get_all(cls, project_id=None) -> List[Self]:
        """
        Method to get all fields belonging to one project
        If called on an instance of the class, the id of the class is used
        :param str project_id: Optionally to get fields from another project as the furthrmind sdk was initiated with, defaults to None
        :return List[Self]: List with instances of field class
        """
        return super().get_all(project_id)
    
    @classmethod
    @BaseClass._create_instances_decorator
    def create(cls, name, type, project_id=None) -> Self:
        """
        Method to create a new sample

        :param name: the name of the field to be created
        :param type: field type of the field. Must be out of:
            - Numeric
            - Date
            - SingleLine
            - ComboBoxEntry
            - MultiLine
            - CheckBox
            - Calculation

        :param project_id: Optionally to create an item in another project as the furthrmind sdk was initiated with
        :return instance of the sample class

        """

        if not name:
            raise ValueError("Name cannot be empty")

        if not type ["Numeric", "Date", "SingleLine", "ComboBoxEntry", "MultiLine", "CheckBox", "Calculation"]:
            raise ValueError("type must be one of Numeric, Date, SingleLine, ComboBoxEntry, MultiLine, CheckBox, Calculation")

        data = {"name": name, "type": type}
        id = cls.post(data, project_id)
        data["id"] = id
        return data

    @classmethod
    @BaseClass._create_instances_decorator
    def create_many(cls, data_list: List[Dict], project_id=None) -> Self:
        """
        Method to create multiple samples

        :param data_list: dict with the following keys:
        - name: the name of the field to be created
        - type: field type of the field. Must be out of:
            - Numeric
            - Date
            - SingleLine
            - ComboBoxEntry
            - MultiLine
            - CheckBox
            - Calculation

        :param project_id: Optionally to create an item in another project as the furthrmind sdk was initiated with
        :return list with instance of the sample class
        """

        for data in data_list:
            if not "name" in data:
                raise ValueError("Name cannot be empty")

            if not data.get("type") in ["Numeric", "Date", "SingleLine", "ComboBoxEntry", "MultiLine", "CheckBox", "Calculation"]:
                raise ValueError(
                    "type must be one of Numeric, Date, SingleLine, ComboBoxEntry, MultiLine, CheckBox, Calculation")

        id_list = cls.post(data_list, project_id)
        for data, id in zip(data_list, id_list):
            data["id"] = id
        return data_list



from furthrmind.collection.baseclass import BaseClassWithFieldData, BaseClass
from typing_extensions import List, Dict, Self, TYPE_CHECKING
from inspect import isclass
if TYPE_CHECKING:
    from furthrmind.collection import *


class Group(BaseClassWithFieldData):
    id = ""
    name = ""
    neglect = False
    shortid = ""
    files: List["File"] = []
    fielddata: List["FieldData"] = []
    experiments: List["Experiment"] = []
    samples: List["Sample"] = []
    researchitems: Dict[str, List["ResearchItem"]] = {}
    sub_groups: List[Self] = []
    parent_group: Self = None

    _attr_definition = {
        "files": {"class": "File"},
        "fielddata": {"class": "FieldData"},
        "samples": {"class": "Sample"},
        "experiments": {"class": "Experiment"},
        "researchitems": {"class": "ResearchItem", "nested_dict": True},
        "sub_groups": {"class": "Group"},
        "parent_group": {"class": "Group"}
    }

    def __init__(self, id=None, data=None):
        super().__init__(id, data)

    def _get_url_instance(self, project_id=None):
        project_url = Group.fm.get_project_url(project_id)
        url = f"{project_url}/groups/{self.id}"
        return url

    @classmethod
    def _get_url_class(cls, id, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/groups/{id}"
        return url

    @classmethod
    def _get_all_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/groups"
        return url

    @classmethod
    def _post_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/groups"
        return url

    @classmethod
    def get(cls, id=None, name=None) -> Self:
        """
        Method to get all one group by it's id or short_id
        If called on an instance of the class, the id of the class is used
        :param str id: id or short_id of requested group 
        :param str name: name of requested group 
        :return Self: Instance of group class
        """


        if isclass(cls):
            if id is None:
                id = name
            assert id is not None or name is not None, "Either id or name must be specified"
            return cls._get_class_method(id)
        else:
            self = cls
            data = self._get_instance_method()
            return data
    
    @classmethod
    def get_all(cls, project_id=None) -> List[Self]:
        """
        Method to get all groups belonging to one project
        :param str project_id: Optionally to get groups from another project as the furthrmind sdk was initiated with, defaults to None
        :return List[Self]: List with instances of group class
        """
        return super().get_all(project_id)
    
    @classmethod
    @BaseClass._create_instances_decorator
    def create(cls, name, project_id=None) -> Self:
        """
        Method to create a new sample
        :param name: the name of the item to be created
        :param project_id: Optionally to create an item in another project as the furthrmind sdk was initiated with
        :return instance of the group class
        """
        data = {"name": name}
        id = cls.post(data, project_id)
        data["id"] = id
        return data

    @classmethod
    @BaseClass._create_instances_decorator
    def create_many(cls, name_list: List[str], project_id=None) -> List[Self]:
        """
        Method to create multiple groups
        :param name_list: list with names of the groups to be created
        :param project_id: Optionally to create an item in another project as the furthrmind sdk was initiated with
        :return list with instance of the group class
        """
        data_list = [{"name": name} for name in name_list]

        id_list = cls.post(data_list, project_id)

        for data, id in zip(data_list, id_list):
            data["id"] = id

        return data_list





from furthrmind.collection.baseclass import BaseClass
import os
from io import BytesIO

class File(BaseClass):
    id = ""
    name = ""

    _attr_definition = {
    }

    def __init__(self, id=None, data=None):
        super().__init__(id, data)

    @classmethod
    def get(cls, id=None):
        raise TypeError("Not implemented")

    @classmethod
    def get_all(cls):
        raise TypeError("Not implemented")


    def download(self, folder, overwrite=False):
        """
        Method to download a file
        :param folder: the folder where the file should be saved
        """
        from furthrmind.file_loader import FileLoader
        fl = FileLoader(self.fm.host, self.fm.api_key)

        if not os.path.isdir(folder):
            raise ValueError("Folder does not exist")
        fl.downloadFile(self.id, folder, overwrite)

    def download_bytes(self) -> BytesIO:
        """
        Method to download a file and save to BytesIO object
        :param folder: the folder where the file should be saved
        """
        from furthrmind.file_loader import FileLoader
        fl = FileLoader(self.fm.host, self.fm.api_key)

        flag, bytesIO = fl.downloadFile(self.id, bytesIO=True)
        return bytesIO

    def update_file(self, file_path, file_name=None):
        from furthrmind.file_loader import FileLoader
        fl = FileLoader(self.fm.host, self.fm.api_key)

        if not os.path.isfile(file_path):
            raise ValueError("File does not exist")

        fl.updateFile(self.id, file_path, file_name)







from ..utils import furthr_wrap
from furthrmind.collection.baseclass import BaseClass
from typing_extensions import Self, List, TYPE_CHECKING
from pandas import DataFrame
from inspect import isclass
if TYPE_CHECKING:
    from furthrmind.collection import Column

class DataTable(BaseClass):
    id = ""
    name = ""
    columns: List["Column"] = []


    _attr_definition = {
        "columns": {"class": "Column"}
    }

    def __init__(self, id=None, data=None):
        super().__init__(id, data)

    def get_all(cls, project_id=None) -> List[Self]:
        raise ValueError("Not implemented for datatables")

    def _get_url_instance(self, project_id=None):
        project_url = DataTable.fm.get_project_url(project_id)
        url = f"{project_url}/rawdata/{self.id}"
        return url

    @classmethod
    def _get_url_class(cls, id, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/rawdata/{id}"
        return url

    @classmethod
    def _get_all_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/rawdata"
        return url

    @classmethod
    def _post_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/rawdata"
        return url
    
    @classmethod
    def get(cls, id=None) -> Self:
        """
        Method to get all one datatable by it's id
        If called on an instance of the class, the id of the class is used
        :param str id: id of requested datatable 
        :return Self: Instance of datatable class
        """
        if isclass(cls):
            return cls._get_class_method(id)
        else:
            self = cls
            data = self._get_instance_method()
            return data
    
    @classmethod
    def get_all(cls, project_id=None) -> List[Self]:
        """
        Method to get all datatables belonging to one project
        :param str project_id: Optionally to get datatables from another project as the furthrmind sdk was initiated with, defaults to None
        :return List[Self]: List with instances of datatable class
        """
        return super().get_all(project_id)

    def get_columns(self, column_id_list: List[str]=None, column_name_list:List[str]=None) -> List["Column"]:
        """
        Method to get columns and their data
        If column_id_list and column_name_list are not provided, the method will retrieve all columns belonging
        to the datatable

        :param column_id_list: list of column_ids to retrieve
        :param column_name_list: list of column names to retrieve
        :return: list of column objects

        """

        columns = self._get_columns(column_id_list, column_name_list)
        new_column_mapping = {c.id: c for c in columns}
        new_column_list = []
        for column in self.columns:
            if column.id in new_column_mapping:
                new_column_list.append(new_column_mapping[column.id])
            else:
                new_column_list.append(column)
        self.columns = new_column_list
        return columns

    def get_pandas_dataframe(self, column_id_list: List[str]=None, column_name_list:List[str]=None) -> DataFrame:
        """
        Method to get columns and their data as a pandas dataframe
        If column_id_list and column_name_list are not provided, the method will retrieve all columns belonging
        to the datatable

        :param column_id_list: list of column_ids to retrieve
        :param column_name_list: list of column names to retrieve
        :return: pandas dataframe

        """

        columns = self._get_columns(column_id_list, column_name_list)
        data_dict = {}
        for c in columns:
            data_dict[c.name] = c.values
        df = DataFrame.from_dict(data_dict)
        return df

    def _get_columns(self, column_id_list: List[str]=None, column_name_list:List[str]=None) -> List["Column"]:
        from furthrmind.collection import Column
        if column_id_list:
            pass
        elif column_name_list:
            column_id_list = []
            for column in self.columns:
                if column.name in column_name_list:
                    column_id_list.append(column.id)
        else:
            column_id_list = [c.id for c in self.columns]
        columns = Column.get_many(column_id_list)
        return columns

    @classmethod
    @BaseClass._create_instances_decorator
    def create(cls, name: str = "Data table", experiment_id=None, sample_id=None, researchitem_id=None, columns=None, project_id=None) -> Self:
        """
        Method to create a new datatable

        :param name: name of the datatable
        :param experiment_id: id of the experiment where the datatable belongs to
        :param sample_id: id of the sample where the datatable belongs to
        :param researchitem_id: id of the researchitem where the datatable belongs to
        :param columns: a list of columns that should be added to the datatable. List with dicts with the following keys:
            - name: name of the column
            - type: Type of the column, Either "Text" or "Numeric". Data must fit to type, for Text all data
            will be converted to string and for Numeric all data is converted to float (if possible)
            - data: List of column values, must fit to column_type
            - unit: dict with id or name, or name as string, or id as string
        :param project_id: Optionally to create an item in another project as the furthrmind sdk was initiated with
        :return: instance of datatable class

        """

        from furthrmind.collection import Column

        if not name:
            raise ValueError("Name must be specified")

        if not experiment_id and not sample_id and not researchitem_id:
            raise ValueError("Either experiment_id or sample_id or researchitem_id must be specified")

        column_id_list = []
        if columns:
            columns = Column.create_many(columns)
            column_id_list = [c.id for c in columns]

        data = {"name": name}
        if column_id_list:
            data["columns"] = [{"id": column_id} for column_id in column_id_list]

        if experiment_id:
            data["experiment"] = {"id": experiment_id}

        if sample_id:
            data["sample"] = {"id": sample_id}

        if researchitem_id:
            data["researchitem"] = {"id":researchitem_id}

        id = cls.post(data, project_id)
        data["id"] = id
        return data

